/*
    libut61 - library for parsing UT61 (ES51922) DMM readout data stream
    Copyright (C) 2021  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page:           http://www.repo.hu/projects/ut61
    Version control:        svn://svn.repo.hu/ut61/trunk
    Contact (email & IRC):  http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "ut61.h"
#include "debug.h"

typedef struct {
	const char *unit;   /* unit the value is displayed in */
	double mult;        /* multiplier to get base unit value */
	double max;         /* maximum value in this range in base unit */
	int ints;           /* integer digits on the left */
} range_t;

typedef struct {
	unsigned char code;
	char vbar;  /* 0 or 1: require vbar match; -1: ignore */
	char judge; /* 0 or 1: require judge match; -1: ignore */
	const char *base_unit;
	const char *name;
	const range_t *ranges;
} func_prop_t;

static const range_t rng_volt[8] = {
	{"V",  1,     2.2,  1},
	{"V",  1,     22,   2},
	{"V",  1,     220,  3},
	{"V",  1,     2200, 4},
	{"mV", 0.001, 2200, 3}
};

static const range_t rng_uamp[8] = {
	{"uA",  0.000001,     220.0/1000000.0, 3},
	{"uA",  0.000001,     2200.0/1000000.0, 4}
};

static const range_t rng_auto_amp1[8] = {
	{"A",  1,     220,   3},
	{"A",  1,     2200,  4}
};

static const range_t rng_mamp[8] = {
	{"mA",  0.001,     22.0/1000.0,  2},
	{"mA",  0.001,     220.0/1000.0, 3}
};

static const range_t rng_auto_amp2[8] = {
	{"A",  1,     22,   2},
	{"A",  1,     220,  3}
};

static const range_t rng_22amp[8] = {
	{"A",  1,     22,   2}
};

static const range_t rng_amp_manual[8] = {
	{"A",  1,     2.2,    1},
	{"A",  1,     22,     2},
	{"A",  1,     220,    3},
	{"A",  1,     2200,   4},
	{"A",  1,     22000,  5}
};

static const range_t rng_auto_ohm[8] = {
	{"Ohm",   1,       220,       3},
	{"kOhm",  1000,    2200,      1},
	{"kOhm",  1000,    22000,     2},
	{"kOhm",  1000,    220000,    3},
	{"MOhm",  1000000, 2200000,   1},
	{"MOhm",  1000000, 22000000,  2},
	{"MOhm",  1000000, 220000000, 3}
};

static const range_t rng_auto_freq[8] = {
	{"Hz",   1,       22,        3},
	{"Hz",   1,       220,       4},
	{NULL,   0,       0,         0},
	{"kHz",  1000,    22000,     2},
	{"kHz",  1000,    220000,    3},
	{"MHz",  1000000, 2200000,   1},
	{"MHz",  1000000, 22000000,  2},
	{"MHz",  1000000, 220000000, 3}
};

static const range_t rng_auto_duty[8] = {
	{"%",    1,       100,       4}
};

static const range_t rng_auto_cap[8] = {
	{"nF",   0.000000001,  22.0/1000000000.0,    2},
	{"nF",   0.000000001,  220.0/1000000000.0,   3},
	{"uF",   0.000001,     2.2/1000000.0,        1},
	{"uF",   0.000001,     22/1000000.0,         2},
	{"uF",   0.000001,     220/1000000.0,        3},
	{"mF",   0.001,        2.2/1000.0,           1},
	{"mF",   0.001,        22/1000.0,            2},
	{"mF",   0.001,        220/1000.0,           3}
};

static const range_t rng_auto_temp[8] = { /* TODO: this has not been tested */
	{"C",    1,       2200,     2},
};

static const range_t rng_unitless[] = {
	{"", 1, 1, 1}
};

#define X -1
#define BIN__(s) (((s[0]-'0') << 6) | ((s[1]-'0') << 5) | ((s[2]-'0') << 4) | ((s[3]-'0') << 3) | ((s[4]-'0') << 2) | ((s[5]-'0') << 1) | (s[6]-'0'))
#define BIN_(s) BIN__((s))
#define BIN(v) BIN_(#v)

static const func_prop_t func_props[] = {
	{BIN(0110010), X, 0, "Hz",  "Frequency", rng_auto_freq},   /* needs to be the first entry for VAHZ */
	{BIN(0110010), X, 1, "%",   "Duty cycle", rng_auto_duty},  /* needs to be the second entry for VAHZ */
	{BIN(0111011), X, X, "V",   "Voltage", rng_volt},
	{BIN(0111101), 0, X, "uA",  "Current (Auto uA)", rng_uamp},
	{BIN(0111101), 1, X, "A",   "Current (Auto 220/2200A)", rng_auto_amp1},
	{BIN(0111111), 0, X, "mA",  "Current (Auto mA)", rng_mamp},
	{BIN(0111111), 1, X, "A",   "Current (Auto 22/220A)", rng_auto_amp2},
	{BIN(0110000), X, X, "A",   "Current (22A)", rng_22amp},
	{BIN(0111001), X, X, "A",   "Current (manual)", rng_amp_manual},
	{BIN(0110011), X, X, "Ohm", "Resistance", rng_auto_ohm},
	{BIN(0110101), X, X, "",    "Continuity", rng_unitless},
	{BIN(0110001), X, X, "",    "Diode", rng_unitless},
	{BIN(0110110), X, X, "F",   "Capacitance", rng_auto_cap},
	{BIN(0110100), X, X, "C",   "Temperature", rng_auto_temp},
	{BIN(0111110), X, X, "",    "ADP", NULL},
	{0}
};

#undef X
#undef BIN

int ut61_digest(ut61_packet_t *pkt)
{
	const func_prop_t *p;
	const range_t *r = NULL;
	static const double shift[10] = { 10000, 1000, 100, 10, 1, 0.1, 0.01, 0.001, 0.0001, 0.00001 };
	int sh, n, m;
	char *s;

	pkt->valid = 0;
	pkt->value = pkt->base_value = 0;
	pkt->unit = pkt->base_unit = pkt->func_name = pkt->info = "<invalid>";
	*pkt->value_str = '\0';

	if (!pkt->raw_valid)
		return -1;

	if (!pkt->vahz) {
		/* normal lookup using the func table */
		for(p = func_props; p->code != 0; p++) {
			if (p->code == pkt->func) {
				if (((p->vbar == -1) || (p->vbar == pkt->vbar)) && ((p->judge == -1) || (p->judge == pkt->judge)))
					break;
			}
		}
	}
	else
		p = &(func_props[pkt->judge]); /* in VAHZ mode we are measuring frequency/duty */

	if (p->code == 0) {
		tprintf(2, "Synatx error: unknonw code %02x\n", p->code);
		return -1;
	}

	pkt->func_name = p->name;

	if (p->ranges == NULL) {
		tprintf(2, "Synatx error: code %02x has no range, can't convert\n", p->code);
		return -1;
	}
	else
		r = &(p->ranges[pkt->range & 0x07]);

/* refuse under/overflow */
	if (pkt->ol) {
		pkt->info = "ol";
		return -1;
	}
	if (pkt->ul) {
		pkt->info = "ul";
		return -1;
	}
	if (pkt->batt)
		pkt->info = "batt";
	else
		pkt->info = "ok";

	/* digit shift for conversion from the range */
	sh = 5 - r->ints;
	if ((sh < 0) || (sh >= 5)) {
		tprintf(2, "Conversion error: invalid shift %d for ints %d\n", sh, r->ints);
		return -1;
	}

	/* convert value */
	pkt->value = 0;
	for(n = 4; n >= 0; n--,sh++)
		pkt->value += (pkt->dig[n] - '0') * shift[sh];

	if (pkt->sign)
		pkt->value = -pkt->value;

	pkt->unit = r->unit;

	/* convert to base value */
	if (r->mult != 0)
		pkt->base_value = pkt->value * r->mult;

	pkt->base_unit = p->base_unit;

	/* convert string value and related fields */
	pkt->dig_left = r->ints;
	pkt->dig_right = 5 - r->ints;

	s = pkt->value_str;
	*s = (pkt->sign) ? '-' : ' ';
	s++;

	for(n = 4, m = r->ints; n >= 0; n--,s++,m--)
	{
		if (m == 0) {
			*s = '.';
			s++;
		}
		if ((pkt->dig[n] == '0') && (m > 1))
			*s = ' '; /* strip leading zeroes, but keep it padded */
		else
			*s = pkt->dig[n];
	}
	*s = '\0';

	pkt->valid = 1;
	return 0;
}

typedef struct {
	FILE *f;
	char *str;
	int maxlen;
} print_t;

#define OUT(chr) \
	do { \
		if (p->str != NULL) { \
			if (p->maxlen == 0) \
				goto out; \
			*p->str = chr; \
			p->str++; \
			p->maxlen--; \
		} \
		else if (p->f != NULL) \
			fputc(chr, p->f); \
		res++; \
	} while(0)

#define S4I(a,b,c,d) \
	((((unsigned long)a) << 24) | (((unsigned long)b) << 16) | (((unsigned long)c) << 8) | (((unsigned long)d)))

#define PBIT(bit) \
	do { \
		buf = buf_; buf_[0] = bit ? '1' : '0'; buf_[1] = '\0'; \
	} while(0)

#define PCHR(chr) \
	do { \
		buf = buf_; buf_[0] = chr; buf_[1] = '\0'; \
	} while(0)

#define PTIME(fmt, field) \
	do { \
		if (tm == NULL) { \
			time_t t = time(NULL); \
			tm = localtime(&t); \
		} \
		sprintf(buf_, fmt, tm->field); buf = buf_; \
	} while(0)

static int template_print(print_t *p, ut61_packet_t *pkt, const char *templ)
{
	const char *s;
	int res = 0, n;
	struct tm *tm = NULL;

	if ((p->str != NULL) && (p->maxlen < 1))
		return -1;

	for(s = templ; *s != '\0'; s++) {
		if (*s == '$') {
			s++;
			if (*s == '(') { /* decode $() */
				unsigned long cmd = 0;
				const char *b, *buf;
				char buf_[256];

				/* load at most 4 characters (up to the closing paren) into a long */
				s++;
				for(n = 0; n < 4; n++) {
					if (*s == '\0')
						goto out;
					if (*s == ')')
						break;
					cmd <<= 8;
					cmd |= *s;
					s++;
				}

				/* expect closing paren and decode */
				if (*s != ')')
					buf = "<template too long>";
				else switch(cmd) {
					case S4I(0,  'v','a','l'): sprintf(buf_, "%f", pkt->value); buf = buf_; break;
					case S4I('v','b','a','s'): sprintf(buf_, "%f", pkt->base_value); buf = buf_; break;
					case S4I('m','b','a','s'): sprintf(buf_, "%.16f", pkt->base_vmax); buf = buf_; break;
					case S4I('v','l','c','d'): buf = pkt->value_str; break;
					case S4I('u','n','i','t'): buf = pkt->unit; break;
					case S4I('u','b','a','s'): buf = pkt->base_unit; break;
					case S4I('f','u','n','c'): buf = pkt->func_name; break;
					case S4I('i','n','f','o'): buf = pkt->info; break;
					case S4I(0,  'v','l','d'): PBIT(pkt->valid); break;
					case S4I(0,  'V','L','D'): PBIT(pkt->raw_valid); break;
					case S4I('s','y','n','c'): PBIT(pkt->synced); break;
					case S4I('j','u','d','g'): PBIT(pkt->judge); break;
					case S4I('s','i','g','n'): PBIT(pkt->sign); break;
					case S4I('b','a','t','t'): PBIT(pkt->batt); break;
					case S4I(0,  0,  'o','l'): PBIT(pkt->ol); break;
					case S4I(0,  0,  'u','l'): PBIT(pkt->ul); break;
					case S4I(0,  'm','i','n'): PBIT(pkt->min); break;
					case S4I(0,  'm','a','x'): PBIT(pkt->max); break;
					case S4I(0,  'r','e','l'): PBIT(pkt->rel); break;
					case S4I(0,  'r','m','r'): PBIT(pkt->rmr); break;
					case S4I('p','m','i','n'): PBIT(pkt->pmin); break;
					case S4I('p','m','a','x'): PBIT(pkt->pmax); break;
					case S4I(0,  0,  'd','c'): PBIT(pkt->dc); break;
					case S4I(0,  0,  'a','c'): PBIT(pkt->ac); break;
					case S4I('a','u','t','o'): PBIT(pkt->automatic); break;
					case S4I('v','a','h','z'): PBIT(pkt->vahz); break;
					case S4I('v','b','a','r'): PBIT(pkt->vbar); break;
					case S4I('h','o','l','d'): PBIT(pkt->hold); break;
					case S4I(0,  'l','p','f'): PBIT(pkt->lpf); break;
					case S4I('F','U','N','C'): sprintf(buf_, "%02x", pkt->func); buf = buf_; break;
					case S4I('R','A','N','G'): sprintf(buf_, "%02x", pkt->range); buf = buf_; break;
					case S4I('d','i','g','0'): PCHR(pkt->dig[0]); break;
					case S4I('d','i','g','1'): PCHR(pkt->dig[1]); break;
					case S4I('d','i','g','2'): PCHR(pkt->dig[2]); break;
					case S4I('d','i','g','3'): PCHR(pkt->dig[3]); break;
					case S4I('d','i','g','4'): PCHR(pkt->dig[4]); break;
					case S4I('T','s','e','c'): PTIME("%02d", tm_sec); break;
					case S4I('T','m','i','n'): PTIME("%02d", tm_min); break;
					case S4I(0,  'T','h','r'): PTIME("%02d", tm_hour); break;
					case S4I('T','d','a','y'): PTIME("%02d", tm_mday); break;
					case S4I('T','m','o','n'): PTIME("%02d", tm_mon); break;
					case S4I(0,  'T','y','r'): PTIME("%02d", tm_year+1900); break;
					case S4I('T','w','d','y'): PTIME("%02d", tm_wday); break;
					case S4I('T','y','d','y'): PTIME("%02d", tm_yday); break;
					case S4I('T','d','s','t'): PTIME("%d",   tm_isdst); break;
					default: buf = "<unknown template>";
				}

				/* print the output from buf */
				for(b = buf; *b != '\0'; b++)
					OUT(*b);
			}
			else
				OUT(*s);
		}
		else if (*s == '\\') { /* handle a few backslash sequences */
			s++;
			switch(*s) {
				case 'n': OUT('\n'); break;
				case 'r': OUT('\r'); break;
				case 't': OUT('\t'); break;
				default:  OUT(*s); break;
			}
		}
		else
			OUT(*s); /* plain literal */
	}

	out:;

	if (p->str != NULL) {
		if (p->maxlen == 0)
			p->str[-1] = '\0';  /* rather lose the last char but terminate */
		else
			p->str[0] = '\0';
	}

	return res;
}

int ut61_template_fprint(FILE *f, ut61_packet_t *pkt, const char *templ)
{
	print_t p = {0};
	p.f = f;
	return template_print(&p, pkt, templ);
}

int ut61_template_sprint(char *dst, int maxlen, ut61_packet_t *pkt, const char *templ)
{
	print_t p = {0};
	p.str = dst;
	p.maxlen = maxlen;
	return template_print(&p, pkt, templ);
}
